from counterweight.styles.styles import (
    Absolute,
    Border,
    BorderEdge,
    BorderKind,
    CellStyle,
    Color,
    Fixed,
    Flex,
    Inset,
    Margin,
    Padding,
    Relative,
    Span,
    Style,
    Typography,
)

__all__ = [
    "Absolute",
    "Border",
    "BorderEdge",
    "BorderKind",
    "CellStyle",
    "Color",
    "Fixed",
    "Flex",
    "Inset",
    "Margin",
    "Padding",
    "Relative",
    "Span",
    "Style",
    "Typography",
]



from __future__ import annotations

from asyncio import Task
from collections.abc import Callable
from typing import ClassVar, Literal, TypeVar

from pydantic import ConfigDict, Field

from counterweight._context_vars import current_event_queue, current_hook_idx
from counterweight.events import StateSet
from counterweight.hooks.types import Deps, Getter, Ref, Setter, Setup
from counterweight.types import ForbidExtras


class UseState(ForbidExtras):
    type: Literal["state"] = "state"
    value: object


class UseRef(ForbidExtras):
    type: Literal["ref"] = "ref"
    ref: Ref[object]


class UseEffect(ForbidExtras):
    type: Literal["effect"] = "effect"
    setup: Setup
    deps: Deps
    new_deps: Deps
    task: Task[None] | None = None

    model_config: ClassVar[ConfigDict] = {
        "arbitrary_types_allowed": True,
    }


T = TypeVar("T")


class InconsistentHookExecution(Exception):
    pass


class Hooks(ForbidExtras):
    data: list[UseState | UseRef | UseEffect] = Field(default_factory=list)

    def use_state(self, initial_value: Getter[T] | T) -> tuple[T, Setter[T]]:
        try:
            hook = self.data[current_hook_idx.get()]
            if not isinstance(hook, UseState):
                raise InconsistentHookExecution(
                    f"Expected a {UseState.__name__} hook, but got a {type(hook).__name__} hook instead."
                )
        except IndexError:
            hook = UseState(value=initial_value() if callable(initial_value) else initial_value)
            self.data.append(hook)

        def set_state(value: T | Callable[[T], T]) -> None:
            if callable(value):
                value = value(hook.value)  # type: ignore[arg-type]
            hook.value = value
            current_event_queue.get().put_nowait(StateSet())

        current_hook_idx.set(current_hook_idx.get() + 1)

        return hook.value, set_state  # type: ignore[return-value]

    def use_ref(self, initial_value: Getter[T] | T) -> Ref[T]:
        try:
            hook = self.data[current_hook_idx.get()]
            if not isinstance(hook, UseRef):
                raise InconsistentHookExecution(
                    f"Expected a {UseRef.__name__} hook, but got a {type(hook).__name__} hook instead."
                )
        except IndexError:
            hook = UseRef(ref=Ref[object](current=initial_value() if callable(initial_value) else initial_value))
            self.data.append(hook)

        current_hook_idx.set(current_hook_idx.get() + 1)

        return hook.ref  # type: ignore[return-value]

    def use_effect(self, setup: Setup, deps: Deps) -> None:
        try:
            hook = self.data[current_hook_idx.get()]
            if not isinstance(hook, UseEffect):
                raise InconsistentHookExecution(
                    f"Expected a {UseEffect.__name__} hook, but got a {type(hook).__name__} hook instead."
                )
        except IndexError:
            hook = UseEffect(
                setup=setup,
                deps=(object(),),  # these deps will never equal anything else
                new_deps=deps,
            )
            self.data.append(hook)

        hook.setup = setup  # we must capture the new setup function to update its closure
        hook.new_deps = deps  # ... but the decision about whether to actually rerun it will be made based on its deps

        current_hook_idx.set(current_hook_idx.get() + 1)

        return None



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field
from functools import lru_cache
from itertools import product
from typing import NamedTuple


class Position(NamedTuple):
    x: int
    y: int

    @classmethod
    @lru_cache(maxsize=2**14)
    def flyweight(cls, x: int, y: int) -> Position:
        return cls(x, y)


@dataclass(slots=True)
class Rect:
    x: int = field(default=0)
    y: int = field(default=0)
    width: int = field(default=0)
    height: int = field(default=0)

    def expand_by(self, edge: Edge) -> Rect:
        return Rect(
            x=self.x - edge.left,
            y=self.y - edge.top,
            width=self.width + edge.left + edge.right,
            height=self.height + edge.top + edge.bottom,
        )

    def x_range(self) -> range:
        return range(self.x, self.x + self.width)

    def y_range(self) -> range:
        return range(self.y, self.y + self.height)

    def xy_range(self) -> Iterator[tuple[int, int]]:
        return product(self.x_range(), self.y_range())

    @property
    def left(self) -> int:
        return self.x

    @property
    def right(self) -> int:
        return self.x + self.width - 1

    @property
    def top(self) -> int:
        return self.y

    @property
    def bottom(self) -> int:
        return self.y + self.height - 1

    def left_edge(self) -> tuple[Position, ...]:
        left = self.left
        return tuple(Position.flyweight(left, y) for y in self.y_range())

    def right_edge(self) -> tuple[Position, ...]:
        right = self.right
        return tuple(Position.flyweight(right, y) for y in self.y_range())

    def top_edge(self) -> tuple[Position, ...]:
        top = self.top
        return tuple(Position.flyweight(x, top) for x in self.x_range())

    def bottom_edge(self) -> tuple[Position, ...]:
        bottom = self.bottom
        return tuple(Position.flyweight(x, bottom) for x in self.x_range())

    def __contains__(self, item: object) -> bool:
        if isinstance(item, Position):
            return item.x in self.x_range() and item.y in self.y_range()
        else:
            return False


@dataclass(slots=True)
class Edge:
    left: int = field(default=0)
    right: int = field(default=0)
    top: int = field(default=0)
    bottom: int = field(default=0)



from __future__ import annotations

from typing import TypeVar, overload

from counterweight._context_vars import current_hook_state
from counterweight.hooks.types import Deps, Getter, Ref, Setter, Setup

T = TypeVar("T")


@overload
def use_state(initial_value: Getter[T]) -> tuple[T, Setter[T]]:
    ...


@overload
def use_state(initial_value: T) -> tuple[T, Setter[T]]:
    ...


def use_state(initial_value: Getter[T] | T) -> tuple[T, Setter[T]]:
    """
    Parameters:
        initial_value: The initial value of the state.
            It can either be the initial value itself, or a zero-argument function that returns the initial value.

    Returns:
        The current value of the state (i.e., for the current render cycle).

        A function that can be called to update the value of the state (e.g., in an event handler).
            It can either be called with the new value of the state,
            or a function that takes the current value of the state and returns the new value of the state.
    """
    return current_hook_state.get().use_state(initial_value)


def use_ref(initial_value: Getter[T] | T) -> Ref[T]:
    """
    Parameters:
        initial_value: the initial value of the ref.
            It can either be the initial value itself, or a zero-argument function that returns the initial value.

    Returns:
        A [`Ref`][counterweight.hooks.Ref] that holds a reference to the given value.
    """
    return current_hook_state.get().use_ref(initial_value)


def use_effect(setup: Setup, deps: Deps | None = None) -> None:
    """
    Parameters:
        setup: The setup function that will be called when the component first mounts
            or if its dependencies have changed (see below).

        deps: The dependencies of the effect.
            If any of the dependencies change, the previous invocation of the `setup` function will be cancelled
            and the `setup` function will be run again.
            If `None`, the `setup` function will be run on every render.
    """
    return current_hook_state.get().use_effect(setup, deps)



from dataclasses import dataclass
from typing import Union

from counterweight.geometry import Position


@dataclass(frozen=True, slots=True)
class _Event:
    pass


@dataclass(frozen=True, slots=True)
class TerminalResized(_Event):
    pass


@dataclass(frozen=True, slots=True)
class KeyPressed(_Event):
    key: str


@dataclass(frozen=True, slots=True)
class MouseMoved(_Event):
    position: Position
    button: int | None


@dataclass(frozen=True, slots=True)
class MouseDown(_Event):
    position: Position
    button: int


@dataclass(frozen=True, slots=True)
class MouseUp(_Event):
    position: Position
    button: int


@dataclass(frozen=True, slots=True)
class StateSet(_Event):
    pass


@dataclass(frozen=True, slots=True)
class Dummy(_Event):
    pass


AnyEvent = Union[
    TerminalResized,
    KeyPressed,
    MouseMoved,
    MouseDown,
    MouseUp,
    StateSet,
    Dummy,
]



from __future__ import annotations

from collections.abc import Generator, Mapping
from enum import Enum
from string import printable

from parsy import Parser, char_from, generate, match_item
from structlog import get_logger

from counterweight.events import AnyEvent, KeyPressed, MouseDown, MouseMoved, MouseUp
from counterweight.geometry import Position

logger = get_logger()


class Key(str, Enum):
    """
    Derived from https://github.com/Textualize/textual/blob/c966243b059f0352e2a23b9695776838195364a3/src/textual/keys.py
    """

    Escape = "escape"  # Also Control-[
    ShiftEscape = "shift+escape"

    ControlA = "ctrl+a"
    ControlB = "ctrl+b"
    ControlC = "ctrl+c"
    ControlD = "ctrl+d"
    ControlE = "ctrl+e"
    ControlF = "ctrl+f"
    ControlG = "ctrl+g"
    ControlH = "ctrl+h"
    ControlI = "ctrl+i"  # Tab
    ControlJ = "ctrl+j"  # Newline
    ControlK = "ctrl+k"
    ControlL = "ctrl+l"
    ControlM = "ctrl+m"  # Carriage return
    ControlN = "ctrl+n"
    ControlO = "ctrl+o"
    ControlP = "ctrl+p"
    ControlQ = "ctrl+q"
    ControlR = "ctrl+r"
    ControlS = "ctrl+s"
    ControlT = "ctrl+t"
    ControlU = "ctrl+u"
    ControlV = "ctrl+v"
    ControlW = "ctrl+w"
    ControlX = "ctrl+x"
    ControlY = "ctrl+y"
    ControlZ = "ctrl+z"

    Left = "left"
    Right = "right"
    Up = "up"
    Down = "down"
    End = "end"
    Insert = "insert"
    Delete = "delete"

    ControlLeft = "ctrl+left"
    ControlRight = "ctrl+right"
    ControlUp = "ctrl+up"
    ControlDown = "ctrl+down"
    ControlHome = "ctrl+home"
    ControlEnd = "ctrl+end"
    ControlInsert = "ctrl+insert"
    ControlDelete = "ctrl+delete"
    ControlPageUp = "ctrl+pageup"
    ControlPageDown = "ctrl+pagedown"

    ShiftLeft = "shift+left"
    ShiftRight = "shift+right"
    ShiftUp = "shift+up"
    ShiftDown = "shift+down"
    ShiftHome = "shift+home"
    ShiftEnd = "shift+end"
    ShiftInsert = "shift+insert"
    ShiftDelete = "shift+delete"
    ShiftPageUp = "shift+pageup"
    ShiftPageDown = "shift+pagedown"

    ControlShiftLeft = "ctrl+shift+left"
    ControlShiftRight = "ctrl+shift+right"
    ControlShiftUp = "ctrl+shift+up"
    ControlShiftDown = "ctrl+shift+down"
    ControlShiftHome = "ctrl+shift+home"
    ControlShiftEnd = "ctrl+shift+end"
    ControlShiftInsert = "ctrl+shift+insert"
    ControlShiftDelete = "ctrl+shift+delete"
    ControlShiftPageUp = "ctrl+shift+pageup"
    ControlShiftPageDown = "ctrl+shift+pagedown"

    AltDelete = "alt+delete"

    BackTab = "shift+tab"  # shift + tab

    F1 = "f1"
    F2 = "f2"
    F3 = "f3"
    F4 = "f4"
    F5 = "f5"
    F6 = "f6"
    F7 = "f7"
    F8 = "f8"
    F9 = "f9"
    F10 = "f10"
    F11 = "f11"
    F12 = "f12"
    F13 = "f13"
    F14 = "f14"
    F15 = "f15"
    F16 = "f16"
    F17 = "f17"
    F18 = "f18"
    F19 = "f19"
    F20 = "f20"
    F21 = "f21"
    F22 = "f22"
    F23 = "f23"
    F24 = "f24"

    ControlF1 = "ctrl+f1"
    ControlF2 = "ctrl+f2"
    ControlF3 = "ctrl+f3"
    ControlF4 = "ctrl+f4"
    ControlF5 = "ctrl+f5"
    ControlF6 = "ctrl+f6"
    ControlF7 = "ctrl+f7"
    ControlF8 = "ctrl+f8"
    ControlF9 = "ctrl+f9"
    ControlF10 = "ctrl+f10"
    ControlF11 = "ctrl+f11"
    ControlF12 = "ctrl+f12"
    ControlF13 = "ctrl+f13"
    ControlF14 = "ctrl+f14"
    ControlF15 = "ctrl+f15"
    ControlF16 = "ctrl+f16"
    ControlF17 = "ctrl+f17"
    ControlF18 = "ctrl+f18"
    ControlF19 = "ctrl+f19"
    ControlF20 = "ctrl+f20"
    ControlF21 = "ctrl+f21"
    ControlF22 = "ctrl+f22"
    ControlF23 = "ctrl+f23"
    ControlF24 = "ctrl+f24"

    ControlSpace = "ctrl-space"
    Tab = "tab"
    Space = "space"
    Enter = "enter"
    Backspace = "backspace"

    def __repr__(self) -> str:
        return str(self)


SINGLE_CHAR_TRANSFORMS: Mapping[bytes, Key] = {
    b"\x1b": Key.Escape,
    b"\t": Key.Tab,
    b"\n": Key.Enter,
    b" ": Key.Space,
    b"\x00": Key.ControlSpace,
    b"\x01": Key.ControlA,
    b"\x02": Key.ControlB,
    b"\x03": Key.ControlC,
    b"\x04": Key.ControlD,
    b"\x05": Key.ControlE,
    b"\x06": Key.ControlF,
    b"\x07": Key.ControlG,
    b"\x08": Key.Backspace,
    b"\x0B": Key.ControlK,
    b"\x0C": Key.ControlL,
    b"\x0E": Key.ControlN,
    b"\x0F": Key.ControlO,
    b"\x10": Key.ControlP,
    b"\x11": Key.ControlQ,
    b"\x12": Key.ControlR,
    b"\x13": Key.ControlS,
    b"\x14": Key.ControlT,
    b"\x15": Key.ControlU,
    b"\x16": Key.ControlV,
    b"\x17": Key.ControlW,
    b"\x18": Key.ControlX,
    b"\x19": Key.ControlY,
    b"\x1A": Key.ControlZ,
    b"\x7f": Key.Backspace,
}

single_transformable_char = char_from(b"".join((printable.encode("utf-8"), *SINGLE_CHAR_TRANSFORMS.keys())))
decimal_digits = char_from(b"0123456789").many().map(b"".join)


@generate
def single_char() -> Generator[Parser, bytes, AnyEvent]:
    c = yield single_transformable_char

    return KeyPressed(key=SINGLE_CHAR_TRANSFORMS.get(c) or c.decode("utf-8"))


esc = match_item(b"\x1b")
left_bracket = match_item(b"[")


@generate
def escape_sequence() -> Generator[Parser, AnyEvent, AnyEvent]:
    keys = yield esc >> (f1to4 | (left_bracket >> (mouse | two_params | zero_or_one_params)))

    return keys


F1TO4 = {
    b"P": Key.F1,
    b"Q": Key.F2,
    b"R": Key.F3,
    b"S": Key.F4,
}

O_PQRS = match_item(b"O") >> char_from(b"PQRS")


@generate
def f1to4() -> Generator[Parser, bytes, AnyEvent]:
    final = yield O_PQRS

    return KeyPressed(key=F1TO4[final])


left_angle = match_item(b"<")
mM = char_from(b"mM")


@generate
def mouse() -> Generator[Parser, bytes, AnyEvent]:
    # https://www.xfree86.org/current/ctlseqs.html
    # https://invisible-island.net/xterm/ctlseqs/ctlseqs.pdf
    yield left_angle

    buttons_ = yield decimal_digits
    yield semicolon
    x_ = yield decimal_digits
    yield semicolon
    y_ = yield decimal_digits
    m = yield mM

    button_info = int(buttons_)
    x = int(x_) - 1
    y = int(y_) - 1

    pos = Position.flyweight(x=x, y=y)
    moving = button_info & 32
    button = (button_info & 0b11) + 1

    if moving:
        return MouseMoved(position=pos, button=button if button != 4 else None)  # raw 3 is released, becomes 4 above
    elif m == b"m":
        return MouseUp(position=pos, button=button)
    else:  # m == b"M"
        return MouseDown(position=pos, button=button)


CSI_LOOKUP: Mapping[tuple[bytes, ...], str] = {
    # 0 params
    (b"", b"A"): Key.Up,
    (b"", b"B"): Key.Down,
    (b"", b"C"): Key.Right,
    (b"", b"D"): Key.Left,
    (b"", b"F"): Key.End,
    (b"", b"Z"): Key.BackTab,
    # 1 param
    (b"2", b"~"): Key.Insert,
    (b"3", b"~"): Key.Delete,
    (b"11", b"~"): Key.F1,
    (b"12", b"~"): Key.F2,
    (b"13", b"~"): Key.F3,
    (b"14", b"~"): Key.F4,
    (b"15", b"~"): Key.F5,
    # skip 16
    (b"17", b"~"): Key.F6,
    (b"18", b"~"): Key.F7,
    (b"19", b"~"): Key.F8,
    (b"20", b"~"): Key.F9,
    (b"21", b"~"): Key.F10,
    # skip 22
    (b"23", b"~"): Key.F11,
    (b"24", b"~"): Key.F12,
    (b"25", b"~"): Key.F13,
    (b"26", b"~"): Key.F14,
    # skip 27
    (b"28", b"~"): Key.F15,
    (b"29", b"~"): Key.F16,
    # skip 30
    (b"31", b"~"): Key.F17,
    (b"32", b"~"): Key.F18,
    (b"33", b"~"): Key.F19,
    (b"34", b"~"): Key.F20,
    # 2 params
    (b"1", b"2", b"A"): Key.ShiftUp,
    (b"1", b"2", b"B"): Key.ShiftDown,
    (b"1", b"2", b"C"): Key.ShiftRight,
    (b"1", b"2", b"D"): Key.ShiftLeft,
    (b"1", b"5", b"A"): Key.ControlUp,
    (b"1", b"5", b"B"): Key.ControlDown,
    (b"1", b"5", b"C"): Key.ControlRight,
    (b"1", b"5", b"D"): Key.ControlLeft,
    (b"1", b"6", b"A"): Key.ControlShiftUp,
    (b"1", b"6", b"B"): Key.ControlShiftDown,
    (b"1", b"6", b"C"): Key.ControlShiftRight,
    (b"1", b"6", b"D"): Key.ControlShiftLeft,
    (b"3", b"3", b"~"): Key.AltDelete,
    (b"3", b"5", b"~"): Key.ControlDelete,
    (b"3", b"6", b"~"): Key.ControlShiftInsert,
}

final_csi_char = char_from(b"".join(sorted(set(key[-1] for key in CSI_LOOKUP))))
semicolon = match_item(b";")


@generate
def two_params() -> Generator[Parser, bytes, AnyEvent]:
    p1 = yield decimal_digits
    yield semicolon
    p2 = yield decimal_digits
    e = yield final_csi_char

    return KeyPressed(key=CSI_LOOKUP[(p1, p2, e)])


@generate
def zero_or_one_params() -> Generator[Parser, bytes, AnyEvent]:
    p1 = yield decimal_digits  # zero params => b""
    e = yield final_csi_char

    return KeyPressed(key=CSI_LOOKUP[(p1, e)])


@generate
def vt_inputs() -> Generator[Parser, list[AnyEvent], list[AnyEvent]]:
    commands = yield (escape_sequence | single_char).many()
    return commands



from __future__ import annotations

import os
import selectors
import termios
from collections.abc import Callable
from copy import deepcopy
from selectors import DefaultSelector
from threading import Event
from time import perf_counter_ns
from typing import TextIO

from structlog import get_logger

from counterweight.events import AnyEvent
from counterweight.keys import vt_inputs

logger = get_logger()


def read_keys(
    stream: TextIO,
    put_event: Callable[[AnyEvent], None],
    allow: Event,
    waiting: Event,
) -> None:
    selector = DefaultSelector()
    selector.register(stream, selectors.EVENT_READ)

    while True:
        if not allow.is_set():
            waiting.set()
            allow.wait()
            waiting.clear()

        # We're just reading from one file,
        # so we can dispense with the ceremony of actually using the results of the select,
        # other than knowing that it did return something.
        if not selector.select(timeout=1 / 60):
            continue

        start_parsing = perf_counter_ns()
        bytes = os.read(stream.fileno(), 2**10)

        if not bytes:
            continue

        try:
            inputs = vt_inputs.parse(bytes)

            for i in inputs:
                put_event(i)

            logger.debug(
                "Parsed user input",
                inputs=inputs,
                bytes=bytes,
                elapsed_ns=f"{perf_counter_ns() - start_parsing:_}",
            )
        except Exception as e:
            logger.error(
                "Failed to parse input",
                error=repr(e),
                bytes=bytes,
                elapsed_ns=f"{perf_counter_ns() - start_parsing:_}",
            )


LFLAG = 3
CC = 6

TCGetAttr = list[int | list[int | bytes]]


def start_input_control(stream: TextIO) -> TCGetAttr:  # pragma: untestable
    original = termios.tcgetattr(stream)

    modified = deepcopy(original)

    modified[LFLAG] = original[LFLAG] & ~(termios.ECHO | termios.ICANON)
    modified[CC][termios.VMIN] = 1
    modified[CC][termios.VTIME] = 0

    termios.tcsetattr(stream.fileno(), termios.TCSADRAIN, modified)

    return original


def stop_input_control(stream: TextIO, original: TCGetAttr) -> None:  # pragma: untestable
    termios.tcsetattr(stream.fileno(), termios.TCSADRAIN, original)



from __future__ import annotations

import shutil
import sys
from asyncio import CancelledError, Queue, QueueEmpty, Task, TaskGroup, get_running_loop
from collections import deque
from collections.abc import Callable
from itertools import chain, repeat
from signal import SIG_DFL, SIGWINCH, signal
from threading import Event, Thread
from time import perf_counter_ns
from typing import Iterable, TextIO

from structlog import get_logger

from counterweight._context_vars import current_event_queue
from counterweight._utils import drain_queue
from counterweight.border_healing import heal_borders
from counterweight.components import Component, component
from counterweight.controls import AnyControl, Bell, Quit, Screenshot, Suspend, ToggleBorderHealing, _Control
from counterweight.elements import AnyElement, Div
from counterweight.events import (
    AnyEvent,
    Dummy,
    KeyPressed,
    MouseDown,
    MouseMoved,
    MouseUp,
    StateSet,
    TerminalResized,
)
from counterweight.geometry import Position
from counterweight.hooks.impls import UseEffect
from counterweight.input import read_keys, start_input_control, stop_input_control
from counterweight.layout import build_layout_tree_from_concrete_element_tree
from counterweight.logging import configure_logging
from counterweight.output import (
    CLEAR_SCREEN,
    paint_to_instructions,
    start_mouse_tracking,
    start_output_control,
    stop_mouse_tracking,
    stop_output_control,
)
from counterweight.paint import BLANK, Paint, paint_layout, svg
from counterweight.shadow import ShadowNode, update_shadow
from counterweight.styles import Span, Style
from counterweight.styles.styles import Flex

logger = get_logger()


def start_handling_resize_signal(put_event: Callable[[AnyEvent], None]) -> None:
    signal(SIGWINCH, lambda _, __: put_event(TerminalResized()))


def stop_handling_resize_signal() -> None:
    signal(SIGWINCH, SIG_DFL)


async def app(
    root: Callable[[], Component],
    output_stream: TextIO = sys.stdout,
    input_stream: TextIO = sys.stdin,
    headless: bool = False,
    dimensions: tuple[int, int] | None = None,
    autopilot: Iterable[AnyEvent | AnyControl] = (),
) -> None:
    """
    Parameters:
        root: The root [component][counterweight.components.component] of the application.
        output_stream: The stream to which the application will write its output.
        input_stream: The stream from which the application will read its input.
        headless: If `True`, the application will not attempt to read from the input stream or write to the output stream.
            This is primarily useful when combined with the `autopilot` parameter.
        dimensions: The dimensions of the terminal window, in characters.
            If `None`, the dimensions of the terminal window will be used.
            Note that if the terminal is resized, the application will still be notified and will update its dimensions accordingly.
        autopilot: An iterable of events and controls to be processed by the application.
            This is primarily useful for testing or generating screenshots programmatically.
            Note that the autopilot will not be processed until after the initial render cycle,
            and that using the autopilot does not automatically cause the application to quit!
    """
    configure_logging()

    w, h = dimensions or shutil.get_terminal_size()

    @component
    def screen() -> Div:
        return Div(
            children=[root()],
            style=Style(
                layout=Flex(
                    direction="column",
                    justify_children="start",
                    align_children="stretch",
                ),
                span=Span(
                    width=w,
                    height=h,
                ),
            ),
        )

    logger.info("Application starting...")

    event_queue: Queue[AnyEvent] = Queue()
    current_event_queue.set(event_queue)

    loop = get_running_loop()

    def put_event(event: AnyEvent) -> None:
        loop.call_soon_threadsafe(event_queue.put_nowait, event)

    if not headless:
        original = start_input_control(stream=input_stream)

    allow_key_thread = Event()
    allow_key_thread.set()

    waiting_key_thread = Event()

    try:
        if not headless:
            start_handling_resize_signal(put_event=put_event)
            start_output_control(stream=output_stream)
            start_mouse_tracking(stream=output_stream)

            key_thread = Thread(
                target=read_keys,
                args=(
                    input_stream,
                    put_event,
                    allow_key_thread,
                    waiting_key_thread,
                ),
                daemon=True,
            )
            key_thread.start()

        current_paint: Paint = {Position.flyweight(x, y): BLANK for x in range(w) for y in range(h)}
        instructions = paint_to_instructions(paint=current_paint)

        if not headless:
            output_stream.write(instructions)
            output_stream.flush()

        needs_render = True
        shadow = update_shadow(screen(), None)
        active_effects: set[Task[None]] = set()

        should_quit = False
        should_bell = False
        should_screenshot: Screenshot | None = None
        should_suspend: Suspend | None = None

        do_heal_borders = True

        def handle_control(control: AnyControl | None) -> None:
            nonlocal needs_render

            nonlocal should_quit
            nonlocal should_bell
            nonlocal should_screenshot
            nonlocal should_suspend

            nonlocal do_heal_borders

            match control:
                case None:
                    pass
                case Quit():
                    should_quit = True
                case Bell():
                    should_bell = True
                case Screenshot():
                    should_screenshot = control
                case Suspend():
                    should_suspend = control
                    needs_render = True
                case ToggleBorderHealing():
                    do_heal_borders = not do_heal_borders
                    needs_render = True

        mouse_position = Position.flyweight(x=-1, y=-1)

        async with TaskGroup() as tg:
            for ap in chain(autopilot, repeat(None)):
                if should_quit:
                    break

                if should_bell:
                    if not headless:
                        output_stream.write("\a")
                        output_stream.flush()
                    should_bell = False

                if should_screenshot:
                    try:
                        start_screenshot = perf_counter_ns()
                        should_screenshot.handler(svg(current_paint))
                        logger.debug(
                            "Took screenshot",
                            handler=should_screenshot.handler,
                            elapsed_ns=f"{perf_counter_ns() - start_screenshot:_}",
                        )
                    except Exception as ex:
                        logger.error(
                            "Error in screenshot handler",
                            error=repr(ex),
                            handler=should_screenshot.handler,
                            elapsed_ns=f"{perf_counter_ns() - start_screenshot:_}",
                        )

                    should_screenshot = None

                if should_suspend:
                    start_suspend = perf_counter_ns()
                    logger.debug(
                        "Suspending application",
                        handler=should_suspend.handler,
                    )

                    if not headless:
                        stop_mouse_tracking(stream=output_stream)
                        stop_output_control(stream=output_stream)
                        stop_input_control(stream=input_stream, original=original)
                        stop_handling_resize_signal()

                        # Coordinate with the key thread to tell it to pause,
                        # and to actually wait for it to pause.
                        allow_key_thread.clear()
                        waiting_key_thread.wait()

                    try:
                        should_suspend.handler()
                    except Exception as ex:
                        logger.error(
                            "Error in suspend handler",
                            handler=should_suspend.handler,
                            error=repr(ex),
                        )

                    if not headless:
                        original = start_input_control(stream=input_stream)
                        start_handling_resize_signal(put_event=put_event)
                        start_output_control(stream=output_stream)
                        start_mouse_tracking(stream=output_stream)

                        allow_key_thread.set()

                    # TODO: below is the same as a terminal resize, should probably refactor
                    # note: we may have missed resize events while suspended, so we definitely want to get term size here

                    w, h = shutil.get_terminal_size()

                    # start from scratch
                    current_paint = {Position.flyweight(x, y): BLANK for x in range(w) for y in range(h)}
                    instructions = paint_to_instructions(paint=current_paint)
                    if not headless:
                        output_stream.write(CLEAR_SCREEN + instructions)
                        # don't flush here, we don't necessarily need to flush until the next render
                        # probably we can even store this until the next render happens and output it then

                    logger.debug(
                        "Resuming application",
                        handler=should_suspend.handler,
                        elapsed_ns=f"{perf_counter_ns() - start_suspend:_}",
                    )

                    should_suspend = None

                if needs_render:
                    start_render = perf_counter_ns()
                    shadow = update_shadow(screen(), shadow)
                    logger.debug(
                        "Updated shadow tree",
                        elapsed_ns=f"{perf_counter_ns() - start_render:_}",
                    )

                    start_concrete = perf_counter_ns()
                    element_tree = build_concrete_element_tree(shadow)
                    logger.debug(
                        "Extracted concrete element tree from shadow tree",
                        elapsed_ns=f"{perf_counter_ns() - start_concrete:_}",
                    )

                    start_layout = perf_counter_ns()
                    layout_tree = build_layout_tree_from_concrete_element_tree(element_tree)
                    layout_tree.compute_layout()
                    logger.debug(
                        "Calculated layout",
                        elapsed_ns=f"{perf_counter_ns() - start_layout:_}",
                    )

                    start_hover = perf_counter_ns()
                    for b in layout_tree.walk_from_bottom():
                        _, border_rect, _ = b.dims.padding_border_margin_rects()
                        if mouse_position in border_rect:
                            # TODO: hover changing layout doesn't really make sense here...
                            b.element = b.element.model_copy(update={"style": b.element.style | b.element.on_hover})
                    logger.debug(
                        "Applied hover styles",
                        elapsed_ns=f"{perf_counter_ns() - start_hover:_}",
                    )

                    start_paint = perf_counter_ns()
                    new_paint = paint_layout(layout_tree)
                    logger.debug(
                        "Generated new paint",
                        elapsed_ns=f"{perf_counter_ns() - start_paint:_}",
                    )

                    if do_heal_borders:
                        start_border_join = perf_counter_ns()
                        new_paint |= heal_borders(new_paint)
                        logger.debug(
                            "Healed borders in new paint",
                            elapsed_ns=f"{perf_counter_ns() - start_border_join:_}",
                        )

                    start_diff = perf_counter_ns()
                    diff = diff_paint(new_paint, current_paint)
                    current_paint |= diff
                    logger.debug(
                        "Diffed new paint from current paint",
                        elapsed_ns=f"{perf_counter_ns() - start_diff:_}",
                        diff_cells=len(diff),
                    )

                    start_instructions = perf_counter_ns()
                    instructions = paint_to_instructions(diff)
                    logger.debug(
                        "Generated instructions from paint diff",
                        elapsed_ns=f"{perf_counter_ns() - start_instructions:_}",
                    )

                    if not headless:
                        start_write = perf_counter_ns()
                        output_stream.write(instructions)
                        output_stream.flush()
                        logger.debug(
                            "Wrote and flushed instructions to output stream",
                            elapsed_ns=f"{perf_counter_ns() - start_write:_}",
                            bytes=f"{len(instructions):_}",
                        )

                    start_effects = perf_counter_ns()
                    active_effects = await handle_effects(shadow, active_effects=active_effects, task_group=tg)
                    logger.debug(
                        "Handled effects",
                        elapsed_ns=f"{perf_counter_ns() - start_effects:_}",
                        num_effects=len(active_effects),
                    )

                    needs_render = False

                    logger.debug("Completed render cycle", elapsed_ns=f"{perf_counter_ns() - start_render:_}")

                if ap is not None:
                    if isinstance(ap, _Control):
                        handle_control(ap)
                        await event_queue.put(Dummy())  # Force a render cycle when the autopilot emits a control
                    else:  # i.e., an event
                        await event_queue.put(ap)
                    logger.debug("Handled autopilot command", command=ap)

                # Goal: wait for an event, then process all pending events, and any events created by those events,
                # until there are no more events immediately available in the queue.
                # This makes sure that (for example) if you use autopilot to press a key which causes a state change,
                # the state change will be processed before the next render cycle.
                # Note that there are some tricky async semantics here. Right now, event handlers are sync,
                # so there's no way for an effect to add events to the queue once we get past the `await drain_queue()` below,
                # since we don't await again until we hit this point again in the next render cycle.

                num_events_handled = 0
                events = deque(await drain_queue(event_queue))

                start_event_handling = perf_counter_ns()
                while events:
                    start_handle_event = perf_counter_ns()

                    event = events.popleft()

                    match event:
                        case StateSet():
                            needs_render = True
                        case TerminalResized():
                            needs_render = True
                            w, h = shutil.get_terminal_size()

                            # start from scratch
                            current_paint = {Position.flyweight(x, y): BLANK for x in range(w) for y in range(h)}
                            instructions = paint_to_instructions(paint=current_paint)
                            if not headless:
                                output_stream.write(CLEAR_SCREEN + instructions)
                                # don't flush here, we don't necessarily need to flush until the next render
                                # probably we can even store this until the next render happens and output it then
                        case KeyPressed():
                            for e in layout_tree.walk_elements_from_bottom():
                                if e.on_key:
                                    handle_control(e.on_key(event))
                        case MouseMoved(position=p):
                            needs_render = True
                            mouse_position = p
                        case MouseDown():
                            for b in layout_tree.walk_from_bottom():
                                _, border_rect, _ = b.dims.padding_border_margin_rects()
                                if mouse_position in border_rect:
                                    if b.element.on_mouse_down:
                                        handle_control(b.element.on_mouse_down(event))
                        case MouseUp():
                            for b in layout_tree.walk_from_bottom():
                                _, border_rect, _ = b.dims.padding_border_margin_rects()
                                if mouse_position in border_rect:
                                    if b.element.on_mouse_up:
                                        handle_control(b.element.on_mouse_up(event))

                    while True:
                        try:
                            events.append(event_queue.get_nowait())
                        except QueueEmpty:
                            break

                    num_events_handled += 1

                    logger.debug(
                        "Handled event",
                        event_obj=event,
                        elapsed_ns=f"{perf_counter_ns() - start_handle_event:_}",
                    )

                logger.debug(
                    "Handled events",
                    num_events=num_events_handled,
                    elapsed_ns=f"{perf_counter_ns() - start_event_handling:_}",
                )

    except (KeyboardInterrupt, CancelledError) as ex:
        logger.debug(f"Caught {ex!r}")
    finally:
        logger.info("Application stopping...")

        if not headless:
            stop_mouse_tracking(stream=output_stream)
            stop_output_control(stream=output_stream)
            stop_input_control(stream=input_stream, original=original)
            stop_handling_resize_signal()

        logger.info("Application stopped")


async def handle_effects(shadow: ShadowNode, active_effects: set[Task[None]], task_group: TaskGroup) -> set[Task[None]]:
    new_effects: set[Task[None]] = set()
    for node in shadow.walk():
        for effect in node.hooks.data:  # TODO: reaching pretty deep here
            if isinstance(effect, UseEffect):
                if effect.deps != effect.new_deps or effect.new_deps is None:
                    effect.deps = effect.new_deps
                    t = task_group.create_task(effect.setup())
                    new_effects.add(t)
                    effect.task = t
                    logger.debug("Created effect", effect=effect)
                else:
                    if effect.task is None:
                        raise Exception("Effect task should never be None at this point")
                    new_effects.add(effect.task)

    for task in active_effects - new_effects:
        task.cancel()
        logger.debug("Cancelled effect task", task=task)

    return new_effects


def build_concrete_element_tree(root: ShadowNode) -> AnyElement:
    return root.element.model_copy(update={"children": [build_concrete_element_tree(child) for child in root.children]})


def diff_paint(new_paint: Paint, current_paint: Paint) -> Paint:
    diff = {}

    for pos, current_cell in current_paint.items():
        new_cell = new_paint.get(pos, BLANK)

        # This looks duplicative, but each of these checks is faster than the next,
        # but less precise, so we can short-circuit earlier on cheaper operations.
        if new_cell is not current_cell and new_cell != current_cell:
            diff[pos] = new_cell

    return diff



from __future__ import annotations

from dataclasses import dataclass, replace
from functools import wraps
from typing import Callable, ParamSpec

from counterweight.elements import AnyElement

P = ParamSpec("P")


def component(func: Callable[P, AnyElement]) -> Callable[P, Component]:
    """
    A decorator that marks a function as a component.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Component:
        return Component(func=func, args=args, kwargs=kwargs)

    return wrapper


@dataclass(frozen=True, slots=True)
class Component:
    """
    The result of calling a component function.
    These should not be instantiated directly;
    instead, use the `@component` decorator on a function
    and call it normally.
    """

    func: Callable[..., AnyElement]
    args: tuple[object, ...]
    kwargs: dict[str, object]
    key: str | int | None = None

    def with_key(self, key: str | int | None) -> Component:
        return replace(self, key=key)



from __future__ import annotations

from typing import Callable, Coroutine, Generic, TypeVar

from pydantic import BaseModel

T = TypeVar("T")

Getter = Callable[[], T]
Setter = Callable[[Callable[[T], T] | T], None]
Setup = Callable[[], Coroutine[None, None, None]]
Deps = tuple[object, ...] | None


class Ref(BaseModel, Generic[T]):
    current: T



from __future__ import annotations

from enum import Enum
from functools import cached_property, lru_cache
from typing import Literal, NamedTuple, TypeVar

from cachetools import LRUCache
from pydantic import Field, NonNegativeInt, PositiveInt

from counterweight.types import FrozenForbidExtras

S = TypeVar("S", bound="StyleFragment")


STYLE_MERGE_CACHE: LRUCache[tuple[int, int], StyleFragment] = LRUCache(maxsize=2**16)


UNSET = object()


def recursive_merge_dicts(a: dict[str, object], b: dict[str, object]) -> dict[str, object]:
    merged: dict[str, object] = {}

    for key in a.keys() | b.keys():
        a_val, b_val = a.get(key, UNSET), b.get(key, UNSET)

        if isinstance(a_val, dict) and isinstance(b_val, dict):
            merged[key] = recursive_merge_dicts(a_val, b_val)
        elif b_val is not UNSET:
            merged[key] = b_val
        elif a_val is not UNSET:
            merged[key] = a_val

    return merged


def merge_style_fragments(left: S, right: S) -> S:
    return type(left).model_validate(
        recursive_merge_dicts(
            left.mergeable_dump(),
            right.mergeable_dump(),
        )
    )


class StyleFragment(FrozenForbidExtras):
    def __or__(self: S, other: S | None) -> S:
        if other is None:
            return self

        key = (
            self._cached_hash,
            other._cached_hash,
        )
        try:
            return STYLE_MERGE_CACHE[key]  # type: ignore[return-value]
        except KeyError:
            merged = merge_style_fragments(self, other)
            STYLE_MERGE_CACHE[key] = merged
            return merged

    @cached_property
    def _cached_hash(self) -> int:
        """This is safe because all style fragments are immutable."""
        return hash(self)

    def mergeable_dump(self) -> dict[str, object]:
        diff: dict[str, object] = {}

        for field_name, value in self:
            field = self.model_fields[field_name]
            field_default = field.default

            if field_name not in self.model_fields_set:
                continue

            if isinstance(value, StyleFragment):
                if isinstance(field_default, StyleFragment):
                    # Value and default are both fragments
                    sub_diff = value.mergeable_dump()
                    if isinstance(field.discriminator, str):
                        sub_diff[field.discriminator] = getattr(value, field.discriminator)
                    if sub_diff:
                        diff[field_name] = sub_diff
                else:
                    # Value is a fragment, default is a primitive
                    diff[field_name] = value.mergeable_dump()
            else:
                # Not a fragment, so it's a primitive
                diff[field_name] = value

        return diff


class Color(NamedTuple):
    red: int
    green: int
    blue: int

    @classmethod
    def from_name(cls, name: str) -> Color:
        return COLORS_BY_NAME[name]

    @classmethod
    @lru_cache(maxsize=2**14)
    def from_hex(cls, hex: str) -> Color:
        hex = hex.lstrip("#")
        return cls(
            int(hex[0:2], 16),
            int(hex[2:4], 16),
            int(hex[4:6], 16),
        )

    @property
    def hex(self) -> str:
        return f"#{self.red:02x}{self.green:02x}{self.blue:02x}"

    def blend(self, other: Color, alpha: float) -> Color:
        return Color(
            red=int(self.red * (1 - alpha) + other.red * alpha),
            green=int(self.green * (1 - alpha) + other.green * alpha),
            blue=int(self.blue * (1 - alpha) + other.blue * alpha),
        )


COLORS_BY_NAME = {
    name: Color.from_hex(hex)
    for name, hex in {
        "aliceblue": "#F0F8FF",
        "antiquewhite": "#FAEBD7",
        "aqua": "#00FFFF",
        "aquamarine": "#7FFFD4",
        "azure": "#F0FFFF",
        "beige": "#F5F5DC",
        "bisque": "#FFE4C4",
        "black": "#000000",
        "blanchedalmond": "#FFEBCD",
        "blue": "#0000FF",
        "blueviolet": "#8A2BE2",
        "brown": "#A52A2A",
        "burlywood": "#DEB887",
        "cadetblue": "#5F9EA0",
        "chartreuse": "#7FFF00",
        "chocolate": "#D2691E",
        "coral": "#FF7F50",
        "cornflowerblue": "#6495ED",
        "cornsilk": "#FFF8DC",
        "crimson": "#DC143C",
        "cyan": "#00FFFF",
        "darkblue": "#00008B",
        "darkcyan": "#008B8B",
        "darkgoldenrod": "#B8860B",
        "darkgray": "#A9A9A9",
        "darkgreen": "#006400",
        "darkkhaki": "#BDB76B",
        "darkmagenta": "#8B008B",
        "darkolivegreen": "#556B2F",
        "darkorange": "#FF8C00",
        "darkorchid": "#9932CC",
        "darkred": "#8B0000",
        "darksalmon": "#E9967A",
        "darkseagreen": "#8FBC8F",
        "darkslateblue": "#483D8B",
        "darkslategray": "#2F4F4F",
        "darkturquoise": "#00CED1",
        "darkviolet": "#9400D3",
        "deeppink": "#FF1493",
        "deepskyblue": "#00BFFF",
        "dimgray": "#696969",
        "dodgerblue": "#1E90FF",
        "firebrick": "#B22222",
        "floralwhite": "#FFFAF0",
        "forestgreen": "#228B22",
        "fuchsia": "#FF00FF",
        "gainsboro": "#DCDCDC",
        "ghostwhite": "#F8F8FF",
        "gold": "#FFD700",
        "goldenrod": "#DAA520",
        "gray": "#808080",
        "green": "#008000",
        "greenyellow": "#ADFF2F",
        "honeydew": "#F0FFF0",
        "hotpink": "#FF69B4",
        "indianred": "#CD5C5C",
        "indigo": "#4B0082",
        "ivory": "#FFFFF0",
        "khaki": "#F0E68C",
        "lavender": "#E6E6FA",
        "lavenderblush": "#FFF0F5",
        "lawngreen": "#7CFC00",
        "lemonchiffon": "#FFFACD",
        "lightblue": "#ADD8E6",
        "lightcoral": "#F08080",
        "lightcyan": "#E0FFFF",
        "lightgoldenrodyellow": "#FAFAD2",
        "lightgreen": "#90EE90",
        "lightgray": "#D3D3D3",
        "lightpink": "#FFB6C1",
        "lightsalmon": "#FFA07A",
        "lightseagreen": "#20B2AA",
        "lightskyblue": "#87CEFA",
        "lightslategray": "#778899",
        "lightsteelblue": "#B0C4DE",
        "lightyellow": "#FFFFE0",
        "lime": "#00FF00",
        "limegreen": "#32CD32",
        "linen": "#FAF0E6",
        "magenta": "#FF00FF",
        "maroon": "#800000",
        "mediumaquamarine": "#66CDAA",
        "mediumblue": "#0000CD",
        "mediumorchid": "#BA55D3",
        "mediumpurple": "#9370DB",
        "mediumseagreen": "#3CB371",
        "mediumslateblue": "#7B68EE",
        "mediumspringgreen": "#00FA9A",
        "mediumturquoise": "#48D1CC",
        "mediumvioletred": "#C71585",
        "midnightblue": "#191970",
        "mintcream": "#F5FFFA",
        "mistyrose": "#FFE4E1",
        "moccasin": "#FFE4B5",
        "navajowhite": "#FFDEAD",
        "navy": "#000080",
        "oldlace": "#FDF5E6",
        "olive": "#808000",
        "olivedrab": "#6B8E23",
        "orange": "#FFA500",
        "orangered": "#FF4500",
        "orchid": "#DA70D6",
        "palegoldenrod": "#EEE8AA",
        "palegreen": "#98FB98",
        "paleturquoise": "#AFEEEE",
        "palevioletred": "#DB7093",
        "papayawhip": "#FFEFD5",
        "peachpuff": "#FFDAB9",
        "peru": "#CD853F",
        "pink": "#FFC0CB",
        "plum": "#DDA0DD",
        "powderblue": "#B0E0E6",
        "purple": "#800080",
        "red": "#FF0000",
        "rosybrown": "#BC8F8F",
        "royalblue": "#4169E1",
        "saddlebrown": "#8B4513",
        "salmon": "#FA8072",
        "sandybrown": "#FAA460",
        "seagreen": "#2E8B57",
        "seashell": "#FFF5EE",
        "sienna": "#A0522D",
        "silver": "#C0C0C0",
        "skyblue": "#87CEEB",
        "slateblue": "#6A5ACD",
        "slategray": "#708090",
        "snow": "#FFFAFA",
        "springgreen": "#00FF7F",
        "steelblue": "#4682B4",
        "tan": "#D2B48C",
        "teal": "#008080",
        "thistle": "#D8BFD8",
        "tomato": "#FF6347",
        "turquoise": "#40E0D0",
        "violet": "#EE82EE",
        "wheat": "#F5DEB3",
        "white": "#FFFFFF",
        "whitesmoke": "#F5F5F5",
        "yellow": "#FFFF00",
        "yellowgreen": "#9ACD32",
    }.items()
}


class CellStyle(StyleFragment):
    foreground: Color = Field(default=Color.from_name("white"))
    background: Color = Field(default=Color.from_name("black"))
    bold: bool = False
    dim: bool = False
    italic: bool = False
    underline: bool = False
    strikethrough: bool = False


class BorderParts(NamedTuple):
    left: str
    right: str
    top: str
    bottom: str
    left_top: str
    right_top: str
    left_bottom: str
    right_bottom: str


# https://www.compart.com/en/unicode/block/U+2500
class BorderKind(Enum):
    Light = BorderParts(
        left="│",
        right="│",
        top="─",
        bottom="─",
        left_top="┌",
        right_top="┐",
        left_bottom="└",
        right_bottom="┘",
    )
    LightRounded = BorderParts(
        left="│",
        right="│",
        top="─",
        bottom="─",
        left_top="╭",
        right_top="╮",
        left_bottom="╰",
        right_bottom="╯",
    )
    LightAngled = BorderParts(
        left="▏",
        right="▕",
        top="▔",
        bottom="▁",
        left_top="/",
        right_top="╲",
        left_bottom="╲",
        right_bottom="/",
    )
    Heavy = BorderParts(
        left="┃",
        right="┃",
        top="━",
        bottom="━",
        left_top="┏",
        right_top="┓",
        left_bottom="┗",
        right_bottom="┛",
    )
    Double = BorderParts(
        left="║",
        right="║",
        top="═",
        bottom="═",
        left_top="╔",
        right_top="╗",
        left_bottom="╚",
        right_bottom="╝",
    )
    Thick = BorderParts(
        left="▌",
        right="▐",
        top="▀",
        bottom="▄",
        left_top="▛",
        right_top="▜",
        left_bottom="▙",
        right_bottom="▟",
    )
    McGugan = BorderParts(  # https://www.willmcgugan.com/blog/tech/post/ceo-just-wants-to-draw-boxes/
        left="▕",
        right="▏",
        top="▁",
        bottom="▔",
        left_top=" ",
        right_top=" ",
        left_bottom=" ",
        right_bottom=" ",
    )
    LightShade = BorderParts(
        left="░",
        right="░",
        top="░",
        bottom="░",
        left_top="░",
        right_top="░",
        left_bottom="░",
        right_bottom="░",
    )
    MediumShade = BorderParts(
        left="▒",
        right="▒",
        top="▒",
        bottom="▒",
        left_top="▒",
        right_top="▒",
        left_bottom="▒",
        right_bottom="▒",
    )
    HeavyShade = BorderParts(
        left="▓",
        right="▓",
        top="▓",
        bottom="▓",
        left_top="▓",
        right_top="▓",
        left_bottom="▓",
        right_bottom="▓",
    )
    Star = BorderParts(
        left="*",
        right="*",
        top="*",
        bottom="*",
        left_top="*",
        right_top="*",
        left_bottom="*",
        right_bottom="*",
    )

    def __repr__(self) -> str:
        return f"BorderKind.{self.name}"


class JoinedBorderParts(NamedTuple):
    vertical: str
    horizontal: str
    left_top: str
    right_top: str
    left_bottom: str
    right_bottom: str
    vertical_right: str
    vertical_left: str
    horizontal_top: str
    horizontal_bottom: str
    horizontal_vertical: str

    def select(self, top: bool, bottom: bool, left: bool, right: bool) -> str | None:
        # This could be a lookup table... not sure if that would be better
        match top, bottom, left, right:
            case True, True, True, True:
                return self.horizontal_vertical
            case True, True, True, False:
                return self.vertical_left
            case True, True, False, True:
                return self.vertical_right
            case True, False, True, True:
                return self.horizontal_top
            case False, True, True, True:
                return self.horizontal_bottom
            case True, True, False, False:
                return self.vertical
            case False, False, True, True:
                return self.horizontal
            case True, False, True, False:
                return self.right_bottom
            case True, False, False, True:
                return self.left_bottom
            case False, True, True, False:
                return self.right_top
            case False, True, False, True:
                return self.left_top
            case _:
                return None

    @property
    def connects_right(self) -> frozenset[str]:
        return frozenset(
            {
                self.horizontal,
                self.left_top,
                self.left_bottom,
                self.horizontal_top,
                self.horizontal_bottom,
                self.vertical_right,
                self.horizontal_vertical,
            }
        )

    @property
    def connects_left(self) -> frozenset[str]:
        return frozenset(
            {
                self.horizontal,
                self.right_top,
                self.right_top,
                self.horizontal_top,
                self.horizontal_bottom,
                self.vertical_left,
                self.horizontal_vertical,
            }
        )

    @property
    def connects_top(self) -> frozenset[str]:
        return frozenset(
            {
                self.vertical,
                self.left_bottom,
                self.right_bottom,
                self.vertical_right,
                self.vertical_left,
                self.horizontal_top,
                self.horizontal_vertical,
            }
        )

    @property
    def connects_bottom(self) -> frozenset[str]:
        return frozenset(
            {
                self.vertical,
                self.left_top,
                self.right_top,
                self.vertical_right,
                self.vertical_left,
                self.horizontal_bottom,
                self.horizontal_vertical,
            }
        )


class JoinedBorderKind(Enum):
    Light = JoinedBorderParts(
        vertical="│",
        horizontal="─",
        left_top="┌",
        right_top="┐",
        left_bottom="└",
        right_bottom="┘",
        vertical_right="├",
        vertical_left="┤",
        horizontal_top="┴",
        horizontal_bottom="┬",
        horizontal_vertical="┼",
    )
    LightRounded = JoinedBorderParts(
        vertical="│",
        horizontal="─",
        left_top="╭",
        right_top="╮",
        left_bottom="╰",
        right_bottom="╯",
        vertical_right="├",
        vertical_left="┤",
        horizontal_top="┴",
        horizontal_bottom="┬",
        horizontal_vertical="┼",
    )
    Heavy = JoinedBorderParts(
        vertical="┃",
        horizontal="━",
        left_top="┏",
        right_top="┓",
        left_bottom="┗",
        right_bottom="┛",
        vertical_right="┣",
        vertical_left="┫",
        horizontal_top="┻",
        horizontal_bottom="┳",
        horizontal_vertical="╋",
    )
    Double = JoinedBorderParts(
        vertical="║",
        horizontal="═",
        left_top="╔",
        right_top="╗",
        left_bottom="╚",
        right_bottom="╝",
        vertical_right="╠",
        vertical_left="╣",
        horizontal_top="╩",
        horizontal_bottom="╦",
        horizontal_vertical="╬",
    )

    def __repr__(self) -> str:
        return f"JoinedBorderKind.{self.name}"


class BorderEdge(Enum):
    Top = "top"
    Bottom = "bottom"
    Left = "left"
    Right = "right"


class Border(StyleFragment):
    kind: BorderKind = Field(default=BorderKind.Light)
    style: CellStyle = Field(default=CellStyle())
    edges: frozenset[BorderEdge] = frozenset({BorderEdge.Top, BorderEdge.Bottom, BorderEdge.Left, BorderEdge.Right})
    contract: int = Field(default=0)


class Margin(StyleFragment):
    top: int = Field(default=0)
    bottom: int = Field(default=0)
    left: int = Field(default=0)
    right: int = Field(default=0)
    color: Color = Field(default=Color.from_name("black"))


class Padding(StyleFragment):
    top: int = Field(default=0)
    bottom: int = Field(default=0)
    left: int = Field(default=0)
    right: int = Field(default=0)
    color: Color = Field(default=Color.from_name("black"))


class Span(StyleFragment):
    width: int | Literal["auto"] = Field(default="auto")
    height: int | Literal["auto"] = Field(default="auto")


class Typography(StyleFragment):
    style: CellStyle = Field(default=CellStyle())
    justify: Literal["left", "center", "right"] = "left"
    wrap: Literal["none", "paragraphs"] = "none"


class Relative(StyleFragment):
    """
    Relative positioning is relative to the parent element's content box.
    Elements occupy space and are laid out next to their siblings according
    to the parent's layout direction.
    """

    type: Literal["relative"] = "relative"
    x: int = 0
    y: int = 0


class Inset(StyleFragment):
    vertical: Literal["top", "center", "bottom"] = "top"
    horizontal: Literal["left", "center", "right"] = "left"


class Absolute(StyleFragment):
    """
    Absolute positioning is relative to the parent element's content box,
    but the element does not occupy space in the layout.

    The `inset` property determines which corner of the
    parent element's content box this element is positioned relative to.
    """

    type: Literal["absolute"] = "absolute"
    x: int = 0
    y: int = 0
    inset: Inset = Field(default=Inset())


class Fixed(StyleFragment):
    """
    Fixed positioning is relative to the screen's top-left corner `(0, 0)`.
    """

    type: Literal["fixed"] = "fixed"
    x: int = 0
    y: int = 0


class Flex(StyleFragment):
    type: Literal["flex"] = "flex"
    direction: Literal["row", "column"] = "row"
    position: Relative | Absolute | Fixed = Field(default=Relative(), discriminator="type")
    weight: PositiveInt | None = 1
    z: int = 0
    align_self: Literal["none", "start", "center", "end", "stretch"] = "none"
    justify_children: Literal[
        "start",
        "center",
        "end",
        "space-between",
        "space-around",
        "space-evenly",
    ] = "start"
    align_children: Literal[
        "start",
        "center",
        "end",
        "stretch",
    ] = "start"
    gap_children: NonNegativeInt = 0


class Style(StyleFragment):
    layout: Flex = Field(default=Flex())
    span: Span = Field(default=Span())
    margin: Margin = Field(default=Margin())
    border: Border | None = Field(default=None)
    padding: Padding = Field(default=Padding())
    typography: Typography = Field(default=Typography())



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from typing import Callable, Literal, Sequence, Union

from pydantic import Field

from counterweight.controls import AnyControl
from counterweight.events import KeyPressed, MouseDown, MouseUp
from counterweight.styles import CellStyle, Style
from counterweight.types import FrozenForbidExtras


class Div(FrozenForbidExtras):
    type: Literal["div"] = "div"
    style: Style = Field(default=Style())
    children: Sequence[Component | AnyElement] = Field(default=())
    on_hover: Style = Field(default=Style())
    on_key: Callable[[KeyPressed], AnyControl | None] | None = None
    on_mouse_down: Callable[[MouseDown], AnyControl | None] | None = None
    on_mouse_up: Callable[[MouseUp], AnyControl | None] | None = None


class Chunk(FrozenForbidExtras):
    content: str
    style: CellStyle = Field(default_factory=CellStyle)

    @cached_property
    def cells(self) -> tuple[CellPaint, ...]:
        return tuple(CellPaint(char=char, style=self.style) for char in self.content)

    @classmethod
    def space(cls) -> Chunk:
        return SPACE

    @classmethod
    def newline(cls) -> Chunk:
        return NEWLINE


SPACE = Chunk(content=" ")
NEWLINE = Chunk(content="\n")


class Text(FrozenForbidExtras):
    type: Literal["text"] = "text"
    content: str | Sequence[Chunk]
    style: Style = Field(default=Style())
    on_hover: Style | None = Field(default=None)
    on_key: Callable[[KeyPressed], AnyControl | None] | None = None
    on_mouse_down: Callable[[MouseDown], AnyControl | None] | None = None
    on_mouse_up: Callable[[MouseUp], AnyControl | None] | None = None

    @property
    def children(self) -> Sequence[Component | AnyElement]:
        return ()

    @cached_property
    def cells(self) -> tuple[CellPaint, ...]:
        if isinstance(self.content, str):
            return tuple(CellPaint(char=char, style=self.style.typography.style) for char in self.content)
        else:
            return sum((chunk.cells for chunk in self.content), ())


AnyElement = Union[
    Div,
    Text,
]

from counterweight.components import Component  # noqa: E402, deferred to avoid circular import

Div.model_rebuild()
Text.model_rebuild()


@dataclass(slots=True)
class CellPaint:
    char: str
    style: CellStyle = field(default=CellStyle())



from __future__ import annotations

from typing import TYPE_CHECKING, TextIO

from structlog import get_logger

from counterweight.geometry import Position
from counterweight.paint import Paint
from counterweight.styles.styles import CellStyle

if TYPE_CHECKING:
    pass
# https://www.xfree86.org/current/ctlseqs.html
# https://invisible-island.net/xterm/ctlseqs/ctlseqs.pdf

CURSOR_ON = "\x1b[?25h"
CURSOR_OFF = "\x1b[?25l"

ALT_SCREEN_ON = "\x1b[?1049h"
ALT_SCREEN_OFF = "\x1b[?1049l"

# 1003 = any event
# 1006 = sgr format
SET_ANY_EVENT_MOUSE_SGR_FORMAT = "\x1b[?1003h\x1b[?1006h"
UNSET_ANY_EVENT_MOUSE_SGR_FORMAT = "\x1b[?1003l\x1b[?1006l"

CLEAR_SCREEN = "\x1b[2J"

BELL = "\x07"

logger = get_logger()


def start_output_control(stream: TextIO) -> None:  # pragma: untestable
    stream.write(ALT_SCREEN_ON)
    stream.write(CURSOR_OFF)
    stream.write(CLEAR_SCREEN)

    stream.flush()


def stop_output_control(stream: TextIO) -> None:  # pragma: untestable
    stream.write(ALT_SCREEN_OFF)
    stream.write(CURSOR_ON)

    stream.flush()


def start_mouse_tracking(stream: TextIO) -> None:  # pragma: untestable
    stream.write(SET_ANY_EVENT_MOUSE_SGR_FORMAT)

    stream.flush()


def stop_mouse_tracking(stream: TextIO) -> None:  # pragma: untestable
    stream.write(UNSET_ANY_EVENT_MOUSE_SGR_FORMAT)

    stream.flush()


def move_to(position: Position) -> str:
    return f"\x1b[{position.y + 1};{position.x + 1}f"


def sgr_from_cell_style(style: CellStyle) -> str:
    fg_r, fg_g, fg_b = style.foreground
    bg_r, bg_g, bg_b = style.background

    sgr = f"\x1b[38;2;{fg_r};{fg_g};{fg_b}m\x1b[48;2;{bg_r};{bg_g};{bg_b}m"

    if style.bold:
        sgr += "\x1b[1m"

    if style.dim:
        sgr += "\x1b[2m"

    if style.italic:
        sgr += "\x1b[3m"

    if style.underline:
        sgr += "\x1b[4m"

    if style.strikethrough:
        sgr += "\x1b[9m"

    return sgr


def paint_to_instructions(paint: Paint) -> str:
    return "".join(f"{move_to(pos)}{sgr_from_cell_style(cell.style)}{cell.char}\x1b[0m" for pos, cell in paint.items())



from __future__ import annotations

from asyncio import Queue, QueueEmpty
from functools import lru_cache
from math import ceil, floor
from typing import List, TypeVar

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


async def drain_queue(queue: Queue[T]) -> List[T]:
    items = [await queue.get()]

    while True:
        try:
            items.append(queue.get_nowait())
        except QueueEmpty:
            break

    return items


@lru_cache(maxsize=2**10)
def halve_integer(x: int) -> tuple[int, int]:
    """Halve an integer, accounting for odd integers by making the first "half" larger by one than the second "half"."""
    half = x / 2
    return ceil(half), floor(half)


@lru_cache(maxsize=2**12)
def partition_int(total: int, weights: tuple[int, ...]) -> list[int]:
    """Partition an integer into a list of integers, with each integer in the list corresponding to the weight at the same index in the weights list."""
    # https://stackoverflow.com/questions/62914824/c-sharp-split-integer-in-parts-given-part-weights-algorithm

    if any(w < 0 for w in weights):
        raise ValueError("Weights must be non-negative")

    if total == 0:  # optimization
        return [0] * len(weights)

    total_weight = sum(weights)

    if not total_weight > 0:
        raise ValueError("Total weight must be positive")

    partition = []
    accumulated_diff = 0.0
    for w in weights:
        exact = total * (w / total_weight)
        rounded = round(exact)
        accumulated_diff += exact - rounded

        if accumulated_diff > 0.5:
            rounded += 1
            accumulated_diff -= 1
        elif accumulated_diff < -0.5:
            rounded -= 1
            accumulated_diff += 1

        partition.append(rounded)

    return partition



from __future__ import annotations

from typing import ClassVar

from pydantic import BaseModel, ConfigDict


class ForbidExtras(BaseModel):
    model_config: ClassVar[ConfigDict] = {
        "extra": "forbid",
    }


class FrozenForbidExtras(ForbidExtras):
    model_config: ClassVar[ConfigDict] = {
        "frozen": True,
    }



from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass
from functools import lru_cache, reduce
from itertools import groupby, product
from operator import itemgetter, or_
from textwrap import dedent
from typing import Literal, assert_never
from xml.etree.ElementTree import Element, ElementTree, SubElement

from structlog import get_logger

from counterweight._utils import halve_integer
from counterweight.elements import AnyElement, CellPaint, Div, Text
from counterweight.geometry import Edge, Position, Rect
from counterweight.layout import LayoutBox, LayoutBoxDimensions, wrap_cells
from counterweight.styles import Border
from counterweight.styles.styles import BorderEdge, CellStyle, Color, Margin, Padding

logger = get_logger()


@dataclass(slots=True)
class P:
    char: str
    style: CellStyle
    z: int

    @classmethod
    @lru_cache(maxsize=2**10)
    def blank(cls, z: int) -> P:
        return cls(
            char=" ",
            style=CellStyle(background=Color.from_name("black")),
            z=z,
        )


BLANK = P.blank(z=-1_000_000)


Paint = dict[Position, P]


def combine_paints(paints: Iterable[Paint]) -> Paint:
    return reduce(or_, paints, {})


def paint_layout(layout: LayoutBox) -> Paint:
    return combine_paints(
        map(
            itemgetter(0),
            sorted(
                (paint_element(l.element, l.dims) for l in layout.walk_from_top()),
                key=lambda pz: pz[1],
            ),
        ),
    )


@lru_cache(maxsize=2**10)
def paint_bg(x_range: range, y_range: range, z: int) -> Paint:
    return {Position.flyweight(x, y): P.blank(z=z) for x, y in product(x_range, y_range)}


def paint_element(element: AnyElement, dims: LayoutBoxDimensions) -> tuple[Paint, int]:
    padding_rect, border_rect, margin_rect = dims.padding_border_margin_rects()

    m = paint_edge(element, element.style.margin, dims.margin, margin_rect)
    b = paint_border(element, element.style.border, border_rect) if element.style.border else {}
    t = paint_edge(element, element.style.padding, dims.padding, padding_rect)

    box = m | b | t

    match element:
        case Div():
            paint = box
        case Text() as e:
            paint = box | paint_text(e, dims.content)
        case _:
            raise NotImplementedError(f"Painting {element} is not implemented")

    # Drawing the background explicitly makes sure that when elements overlaps,
    # the element with lower z is actually hidden and doesn't show through.
    # However, we only want to do this for elements that actually have anything in their paint,
    # so that "anonymous" grouping elements don't hide things behind them.
    return (
        (paint_bg(margin_rect.x_range(), margin_rect.y_range(), element.style.layout.z) | paint) if paint else paint,
        element.style.layout.z,
    )


def justify_line(line: list[CellPaint], width: int, justify: Literal["left", "right", "center"]) -> list[CellPaint]:
    space = width - len(line)
    if space <= 0:
        return line
    elif justify == "left":
        return line + [CellPaint(char=" ")] * space
    elif justify == "right":
        return [CellPaint(char=" ")] * space + line
    elif justify == "center":
        left, right = halve_integer(space)
        return [CellPaint(char=" ")] * left + line + [CellPaint(char=" ")] * right
    else:
        assert_never(justify)


def paint_text(text: Text, rect: Rect) -> Paint:
    style = text.style.typography.style

    paint = {}
    lines = wrap_cells(
        cells=text.cells,
        wrap=text.style.typography.wrap,
        width=rect.width,
    )

    previous_cell_style = None

    for y, line in enumerate(lines[: rect.height], start=rect.y):
        justified_line = justify_line(line, rect.width, text.style.typography.justify)
        for x, cell in enumerate(justified_line[: rect.width], start=rect.x):
            cell_style = cell.style

            # Optimization: reuse the same merged style object if the cell style is the same object.
            # This helps when painting a lot of text with the same style.
            if cell_style is not previous_cell_style:
                merged_style = style | cell_style
                previous_cell_style = cell_style

            paint[Position.flyweight(x, y)] = P(
                char=cell.char,
                style=merged_style,  # merged_style will never be unassigned here, since we know previous_cell_style starts as None
                z=text.style.layout.z,
            )

    return paint


def paint_edge(element: AnyElement, mp: Margin | Padding, edge: Edge, rect: Rect, char: str = " ") -> Paint:
    cell_paint = P(char=char, style=CellStyle(background=mp.color), z=element.style.layout.z)

    chars = {}

    # top
    for y in range(rect.top, rect.top + edge.top):
        for x in rect.x_range():
            chars[Position.flyweight(x, y)] = cell_paint

    # bottom
    for y in range(rect.bottom, rect.bottom - edge.bottom, -1):
        for x in rect.x_range():
            chars[Position.flyweight(x, y)] = cell_paint

    # left
    for x in range(rect.left, rect.left + edge.left):
        for y in rect.y_range():
            chars[Position.flyweight(x, y)] = cell_paint

    # right
    for x in range(rect.right, rect.right - edge.right, -1):
        for y in rect.y_range():
            chars[Position.flyweight(x, y)] = cell_paint

    return chars


def paint_border(element: AnyElement, border: Border, rect: Rect) -> Paint:
    style = border.style

    bv = border.kind.value
    left = bv.left
    right = bv.right
    top = bv.top
    bottom = bv.bottom
    left_top = bv.left_top
    right_top = bv.right_top
    left_bottom = bv.left_bottom
    right_bottom = bv.right_bottom

    contract = border.contract
    chars = {}

    rect_left = rect.left
    rect_top = rect.top
    rect_right = rect.right
    rect_bottom = rect.bottom

    draw_left = BorderEdge.Left in border.edges
    draw_right = BorderEdge.Right in border.edges
    draw_top = BorderEdge.Top in border.edges
    draw_bottom = BorderEdge.Bottom in border.edges

    if contract:
        contract_top = contract if not draw_top else None
        contract_bottom = -contract if not draw_bottom else None
        contract_left = contract if not draw_left else None
        contract_right = -contract if not draw_right else None
    else:
        contract_top = contract_bottom = contract_left = contract_right = None

    v_slice = slice(contract_top, contract_bottom)
    h_slice = slice(contract_left, contract_right)

    if draw_left:
        left_paint = P(char=left, style=style, z=element.style.layout.z)
        for p in rect.left_edge()[v_slice]:
            chars[p] = left_paint

    if draw_right:
        right_paint = P(char=right, style=style, z=element.style.layout.z)
        for p in rect.right_edge()[v_slice]:
            chars[p] = right_paint

    if draw_top:
        top_paint = P(char=top, style=style, z=element.style.layout.z)
        for p in rect.top_edge()[h_slice]:
            chars[p] = top_paint

        if draw_left:
            chars[Position.flyweight(x=rect_left, y=rect_top)] = P(char=left_top, style=style, z=element.style.layout.z)
        if draw_right:
            chars[Position.flyweight(x=rect_right, y=rect_top)] = P(
                char=right_top, style=style, z=element.style.layout.z
            )

    if draw_bottom:
        bottom_paint = P(char=bottom, style=style, z=element.style.layout.z)
        for p in rect.bottom_edge()[h_slice]:
            chars[p] = bottom_paint

        if draw_left:
            chars[Position.flyweight(x=rect_left, y=rect_bottom)] = P(
                char=left_bottom, style=style, z=element.style.layout.z
            )
        if draw_right:
            chars[Position.flyweight(x=rect_right, y=rect_bottom)] = P(
                char=right_bottom, style=style, z=element.style.layout.z
            )

    return chars


def svg(paint: Paint) -> ElementTree:
    w, h = max(paint.keys())

    # Measurements start from the top-left corner of each cell, so the width/height need be 1 unit larger for the actual content
    w += 1
    h += 1

    x_mul = 0.55  # x coordinates get cut roughly in half because monospace cells are twice as tall as they are wide
    y_mul = 1.13  # seems to make border connect up just right

    fmt = "0.2f"
    unit = "em"

    root = Element(
        "svg",
        {
            "xmlns": "http://www.w3.org/2000/svg",
            "width": f"{w * x_mul:{fmt}}{unit}",
            "height": f"{h * y_mul:{fmt}}{unit}",
        },
    )

    style = SubElement(
        root,
        "style",
        {},
    )
    style.text = dedent(
        """\
        svg {
          dominant-baseline: hanging;
          font-size: 1rem;
        }
        """
    )

    background_root = SubElement(
        root,
        "g",
        {
            "shape-rendering": "crispEdges",
        },
    )
    SubElement(  # default background color is black, so do a single black rectangle to cover the whole background as an optimization
        background_root,
        "rect",
        {
            "x": f"{0 * x_mul:{fmt}}{unit}",
            "y": f"{0 * y_mul:{fmt}}{unit}",
            "width": f"{w * x_mul:{fmt}}{unit}",
            "height": f"{h * y_mul:{fmt}}{unit}",
            "fill": Color.from_name("black").hex,
        },
    )

    text_root = SubElement(
        root,
        "text",
        {
            "fill": Color.from_name("white").hex,
            "font-family": "monospace",
        },
    )

    rows = defaultdict(list)
    for pos, cell in sorted(paint.items()):
        rows[pos.y].append((pos.x, cell))

    for y, cells in sorted(rows.items()):
        row_tspan_root = SubElement(
            text_root,
            "tspan",
            {
                "y": f"{y * y_mul:{fmt}}{unit}",
            },
        )

        for bg_color, x_cells_group in groupby(cells, key=lambda x_cell: x_cell[1].style.background.hex):
            # Optimization: write out long horizontal rectangles of the same background color as rectangles instead of individual cell-sized rectangles
            x_cells = tuple(x_cells_group)
            first_x, first_cell = x_cells[0]

            SubElement(
                background_root,
                "rect",
                {
                    "x": f"{first_x * x_mul:{fmt}}{unit}",
                    "y": f"{y * y_mul:{fmt}}{unit}",
                    "width": f"{len(x_cells) * x_mul:{fmt}}{unit}",
                    "height": f"{1 * y_mul:{fmt}}{unit}",
                    "fill": bg_color,
                },
            )

            for x, cell in x_cells:
                if cell.char == " ":  # optimization: don't write spaces
                    continue

                ts = SubElement(
                    row_tspan_root,
                    "tspan",
                    {
                        "x": f"{x * x_mul:{fmt}}{unit}",
                    },
                )

                # optimization: don't write white, it's the default
                if cell.style.foreground != Color.from_name("white"):
                    ts.attrib["fill"] = cell.style.foreground.hex
                ts.text = cell.char

    return ElementTree(element=root)



from logging import NOTSET
from pathlib import Path
from shutil import get_terminal_size
from tempfile import gettempdir
from time import sleep

import structlog
from structlog import WriteLoggerFactory

from counterweight.constants import PACKAGE_NAME

DEVLOG_FILE = Path(gettempdir()) / f"{PACKAGE_NAME}.log"


def configure_logging() -> None:
    DEVLOG_FILE.write_text("")  # truncate file and make sure it exists

    structlog.configure(
        processors=[
            structlog.contextvars.merge_contextvars,
            structlog.processors.add_log_level,
            structlog.processors.StackInfoRenderer(),
            structlog.dev.set_exc_info,
            structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S.%f", utc=False),
            structlog.dev.ConsoleRenderer(),
        ],
        wrapper_class=structlog.make_filtering_bound_logger(NOTSET),
        context_class=dict,
        logger_factory=WriteLoggerFactory(DEVLOG_FILE.open(mode="w")),
        cache_logger_on_first_use=True,
    )


def tail_devlog() -> None:
    while True:
        DEVLOG_FILE.touch(exist_ok=True)
        with DEVLOG_FILE.open(mode="r") as f:
            w, _ = get_terminal_size()
            while True:
                line = f.readline()
                if line:
                    print(line, end="")
                else:
                    if f.tell() > DEVLOG_FILE.stat().st_size:
                        # the file is shorter than our current position, so it was truncated
                        print(" DevLog Rotated ".center(w, "━"))
                        break
                    else:
                        sleep(0.01)



from __future__ import annotations

import sys
from importlib import metadata

PACKAGE_NAME = "counterweight"
__version__ = metadata.version(PACKAGE_NAME)
__python_version__ = ".".join(map(str, sys.version_info))



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field
from itertools import zip_longest

from structlog import get_logger

from counterweight._context_vars import current_hook_idx, current_hook_state
from counterweight.components import Component
from counterweight.elements import AnyElement
from counterweight.hooks.impls import Hooks

logger = get_logger()


@dataclass(slots=True)
class ShadowNode:
    component: Component | None
    element: AnyElement
    hooks: Hooks
    children: list[ShadowNode] = field(default_factory=list)

    def walk(self) -> Iterator[ShadowNode]:
        yield self
        for child in self.children:
            if isinstance(child, ShadowNode):
                yield from child.walk()


def update_shadow(next: Component | AnyElement, previous: ShadowNode | None) -> ShadowNode:
    match next, previous:
        case Component(
            func=next_func,
            args=next_args,
            kwargs=next_kwargs,
            key=next_key,
        ) as next_component, ShadowNode(
            component=previous_component,
            children=previous_children,
            hooks=previous_hooks,
        ) if (
            previous_component is not None
            and next_func == previous_component.func
            and next_key == previous_component.key
        ):
            reset_current_hook_idx = current_hook_idx.set(0)
            reset_current_hook_state = current_hook_state.set(previous_hooks)

            element = next_component.func(*next_args, **next_kwargs)

            children = []
            for new_child, previous_child in zip_longest(element.children, previous_children):
                if new_child is None:
                    continue
                children.append(update_shadow(new_child, previous_child))

            new = ShadowNode(
                component=next_component,
                element=element,
                children=children,
                hooks=previous_hooks,  # the hooks are mutable and carry through renders
            )

            current_hook_idx.reset(reset_current_hook_idx)
            current_hook_state.reset(reset_current_hook_state)

            # logger.debug(
            #     "Updated shadow node",
            #     type="component",
            #     id=id,
            #     generation=new.generation,
            # )
        case Component(func=next_func, args=next_args, kwargs=next_kwargs) as next_component, _:
            reset_current_hook_idx = current_hook_idx.set(0)

            hook_state = Hooks()
            reset_current_hook_state = current_hook_state.set(hook_state)

            element = next_func(*next_args, **next_kwargs)

            children = [update_shadow(child, None) for child in element.children]

            new = ShadowNode(
                component=next_component,
                element=element,
                children=children,
                hooks=hook_state,
            )

            current_hook_idx.reset(reset_current_hook_idx)
            current_hook_state.reset(reset_current_hook_state)
        case element, ShadowNode(
            children=previous_children,
            hooks=previous_hooks,
        ):
            children = []
            for new_child, previous_child in zip_longest(element.children, previous_children):
                if new_child is None:
                    continue
                children.append(update_shadow(new_child, previous_child))

            new = ShadowNode(
                component=None,
                element=element,
                children=children,
                hooks=previous_hooks,  # the hooks are mutable and carry through renders
            )
        case element, None:
            new = ShadowNode(
                component=None,
                element=element,
                children=[update_shadow(child, None) for child in element.children],
                hooks=Hooks(),
            )
        case _:
            raise Exception("Unreachable!")

    return new
